/*
 *  Aethyra
 *  Copyright (C) 2004  The Mana World Development Team
 *
 *  This file is part of Aethyra based on original code
 *  from The Mana World.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <climits>

#include <guichan/exception.hpp>
#include <guichan/focushandler.hpp>
#include <guichan/font.hpp>

#include "container.h"
#include "imagebutton.h"
#include "resizegrip.h"
#include "window.h"

#include "../gui.h"
#include "../layout.h"
#include "../palette.h"
#include "../skin.h"

#include "../../../core/configuration.h"
#include "../../../core/log.h"

#include "../../../core/image/image.h"

#include "../../../core/utils/dtor.h"

int Window::instances = 0;
int Window::mouseResize = 0;

gcn::Widget *Window::mPreviousFocus = NULL;

Window::Window(const std::string& caption, bool modal, Window *parent,
               const std::string& skin, bool visible):
    gcn::Window(caption),
    mOldVisibility(false),
    mGrip(NULL),
    mClose(NULL),
    mParent(parent),
    mLayout(NULL),
    mWindowName("window"),
    mDefaultSkinPath(skin),
    mShowTitle(true),
    mModal(modal),
    mDefaultVisible(visible),
    mSaveVisibility(true),
    mHasMaxDimensions(false),
    mMinWinWidth(100),
    mMinWinHeight(40),
    mMaxWinWidth(graphics->getWidth()),
    mMaxWinHeight(graphics->getHeight())
{
    logger->log("Window::Window(\"%s\")", caption.c_str());

    if (!windowContainer)
        throw GCN_EXCEPTION("Window::Window(): no windowContainer set");

    if (!skinLoader)
        skinLoader = new SkinLoader();

    instances++;

    setFrameSize(0);
    setPadding(3);
    setTitleBarHeight(20);

    // Loads the skin
    mSkin = skinLoader->load(skin, mDefaultSkinPath);

    // Add this window to the window container
    windowContainer->add(this);

    if (mModal)
    {
        gui->setCursorType(Gui::CURSOR_POINTER);
        requestModalFocus();
    }

    // Windows are invisible by default
    setVisible(false);

    addWidgetListener(this);
}

Window::~Window()
{
    logger->log("Window::~Window(\"%s\")", getCaption().c_str());

    saveWindowState();

    destroy(mLayout);
    destroy(mClose);

    while (!mWidgets.empty())
    {
        if (mWidgets.front() == mPreviousFocus)
            mPreviousFocus = NULL;
        destroy(mWidgets.front());
    }

    windowContainer->remove(this);

    removeWidgetListener(this);

    instances--;

    mSkin->instances--;

    if (instances == 0)
        destroy(skinLoader);
}

void Window::requestFocus()
{
    mFocusHandler->focusNone();
}

void Window::draw(gcn::Graphics *graphics)
{
    Graphics *g = static_cast<Graphics*>(graphics);

    g->drawImageRect(0, 0, getWidth(), getHeight(), mSkin->getBorder());

    // Draw title
    if (mShowTitle)
    {
        g->setColor(guiPalette->getColor(Palette::TEXT));
        g->setFont(getFont());
        g->drawText(getCaption(), 7, 5, gcn::Graphics::LEFT);
    }

    drawChildren(graphics);

    // Draw Close Button
    if (mClose)
    {
        graphics->pushClipArea(mClose->getDimension());
        mClose->draw(graphics);
        graphics->popClipArea();
    }
}

void Window::setContentSize(int width, int height)
{
    width = width + 2 * getPadding();
    height = height + getPadding() + getTitleBarHeight();

    if (getMinWidth() > width)
        width = getMinWidth();
    else if (getMaxWidth() < width)
        width = getMaxWidth();
    if (getMinHeight() > height)
        height = getMinHeight();
    else if (getMaxHeight() < height)
        height = getMaxHeight();

    setSize(width, height);
}

void Window::setLocationRelativeTo(gcn::Widget *widget)
{
    int wx, wy;
    int x, y;

    widget->getAbsolutePosition(wx, wy);
    getAbsolutePosition(x, y);

    setPosition(getX() + (wx + (widget->getWidth() - getWidth()) / 2 - x),
                getY() + (wy + (widget->getHeight() - getHeight()) / 2 - y));
}

void Window::setLocationRelativeTo(ImageRect::ImagePosition position,
                                   const int offsetX, const int offsetY)
{
    setLocationRelativeTo(graphics->getWidth(), graphics->getHeight(), position,
                          offsetX, offsetY);
}

void Window::setLocationRelativeTo(const int width, const int height,
                                   ImageRect::ImagePosition position,
                                   const int offsetX, const int offsetY)
{
    int x = 0, y = 0;

    getRelativeOffset(position, x, y, width, height, 0, 0);

    setPosition(x - offsetX, y - offsetY);
}

void Window::getRelativeOffset(ImageRect::ImagePosition position, int &x, int &y,
                               const int width, const int height)
{
    getRelativeOffset(position, x, y, graphics->getWidth(),
                      graphics->getHeight(), width, height);
}

void Window::getRelativeOffset(ImageRect::ImagePosition position, int &x, int &y,
                               const int conWidth, const int conHeight,
                               const int width, const int height)
{
    switch (position)
    {
        case ImageRect::UPPER_CENTER:
        case ImageRect::CENTER:
        case ImageRect::LOWER_CENTER:
            x += (conWidth - width) / 2;
            break;
        case ImageRect::UPPER_RIGHT:
        case ImageRect::RIGHT:
        case ImageRect::LOWER_RIGHT:
            x += conWidth - width;
            break;
        default:
            break;
    }

    switch (position)
    {
        case ImageRect::LEFT:
        case ImageRect::CENTER:
        case ImageRect::RIGHT:
            y += (conHeight - height) / 2;
            break;
        case ImageRect::LOWER_LEFT:
        case ImageRect::LOWER_CENTER:
        case ImageRect::LOWER_RIGHT:
            y += conHeight - height;
            break;
        default:
            break;
    }
}

void Window::saveRelativeLocation(const int x, const int y)
{
    ImageRect::ImagePosition savedPos = ImageRect::UPPER_LEFT;
    int posX = 0, posY = 0, offsetX = 0, offsetY = 0, distX = 0, distY = 0;
    int bestOffsetX = 0, bestOffsetY = 0;
    int bestSquaredDistance = INT_MAX;

    for (int pos = ImageRect::UPPER_LEFT; pos <= ImageRect::LOWER_RIGHT &&
         bestSquaredDistance > 0; pos++)
    {
        int squaredDistance = 0;

        // Get the relative position's coordinates. This will be used as a
        // multiplier to determine centered-ness.
        posX = 0; posY = 0;
        getRelativeOffset((ImageRect::ImagePosition) pos, posX, posY, getWidth(),
                           getHeight(), 0, 0);

        // Now determine the actual offset values to be used.
        offsetX = -x; offsetY = -y;
        getRelativeOffset((ImageRect::ImagePosition) pos, offsetX, offsetY, 0, 0);

        // Compare the two values to get the relative distance between the two.
        // Whichever of the two returns the lowest squared distance will cause
        // the least disruption on changing resolution, and the most likely to
        // match where the user would try to locate it themselves.
        distX = offsetX - posX;
        distY = offsetY - posY;

        squaredDistance = distX * distX + distY * distY;

        if (squaredDistance < bestSquaredDistance)
        {
            savedPos = (ImageRect::ImagePosition) pos;
            bestSquaredDistance = squaredDistance;
            bestOffsetX = offsetX;
            bestOffsetY = offsetY;
        }
    }

    mPosition = savedPos;
    mOffsetX = bestOffsetX;
    mOffsetY = bestOffsetY;

    config.setValue(mWindowName + "Position", savedPos);
    config.setValue(mWindowName + "OffsetX",  bestOffsetX);
    config.setValue(mWindowName + "OffsetY",  bestOffsetY);
}

void Window::adaptToNewSize()
{
    setLocationRelativeTo(mPosition, mOffsetX, mOffsetY);

    if (!mHasMaxDimensions)
    {
        mMaxWinWidth = graphics->getWidth();
        mMaxWinHeight = graphics->getHeight();
    }
}

void Window::setMinWidth(int width)
{
    mMinWinWidth = width > mSkin->getMinWidth() ? width : mSkin->getMinWidth();
}

void Window::setMinHeight(int height)
{
    mMinWinHeight = height > mSkin->getMinHeight() ?
                    height : mSkin->getMinHeight();
}

void Window::setMaxWidth(int width)
{
    mMaxWinWidth = width;
    mHasMaxDimensions = true;
}

void Window::setMaxHeight(int height)
{
    mMaxWinHeight = height;
    mHasMaxDimensions = true;
}

void Window::setResizable(bool r)
{
    if ((bool) mGrip == r)
        return;

    if (r)
    {
        mGrip = new ResizeGrip();
        mGrip->setX(getWidth() - mGrip->getWidth() - getChildrenArea().x);
        mGrip->setY(getHeight() - mGrip->getHeight() - getChildrenArea().y);
        add(mGrip);
    }
    else
    {
        remove(mGrip);
        destroy(mGrip);
    }
}

gcn::Widget *Window::getWidgetAt(int x, int y)
{
    if (mClose && mClose->getDimension().isPointInRect(x, y))
        return mClose;

    Widget *widget = gcn::BasicContainer::getWidgetAt(x, y);
    return widget;
}

void Window::widgetResized(const gcn::Event &event)
{
    const gcn::Rectangle area = getChildrenArea();

    if (mGrip)
        mGrip->setPosition(getWidth() - mGrip->getWidth() - area.x,
                           getHeight() - mGrip->getHeight() - area.y);

    if (mClose)
        mClose->setPosition(getWidth() - getPadding() - mClose->getWidth(),
                            getPadding());

    refreshLayout();
}

void Window::widgetShown(const gcn::Event& event)
{
    mVisible = true;

    requestMoveToTop();
    if (config.getValue("autofocus", 1) == 1)
        requestFocus();

    int widgetX, widgetY;
    getAbsolutePosition(widgetX, widgetY);

    widgetX = gui->getMouseX() - widgetX;
    widgetY = gui->getMouseY() - widgetY;

    if (getChildrenArea().isPointInRect(widgetX, widgetY))
    {
        gcn::MouseEvent mouseEvent(this, false, false, false, false,
                                   gcn::MouseEvent::MOVED, 0, widgetX, widgetY, 0);

        mouseMoved(mouseEvent);
    }
}

void Window::widgetHidden(const gcn::Event& event)
{
    mVisible = false;

    WidgetListIterator it;

    for (it = mWidgets.begin(); it != mWidgets.end(); it++)
    {
        if (mFocusHandler->isFocused(*it))
            mFocusHandler->focusNone();
    }

    int widgetX, widgetY;
    getAbsolutePosition(widgetX, widgetY);

    widgetX = gui->getMouseX() - widgetX;
    widgetY = gui->getMouseY() - widgetY;

    gcn::MouseEvent mouseEvent(this, false, false, false, false,
                               gcn::MouseEvent::EXITED, 0, widgetX, widgetY, 0);

    mouseExited(mouseEvent);
}

void Window::action(const gcn::ActionEvent &event)
{
    if (mClose && event.getId() == "close")
        close();
}

void Window::setCloseButton(bool flag)
{
    if ((bool) mClose == flag)
        return;

    if (flag)
    {
        mClose = new ImageButton(mSkin->getCloseImage(), "close", this);
        mClose->setX(getWidth() - getPadding() - mClose->getWidth());
        mClose->setY(getPadding());

        // This is basically the same as a BasicContainer::add(), which wasn't
        // used so that we can ignore the child container dimensions.
        mClose->_setParent(this);
        mClose->_setFocusHandler(_getFocusHandler());
    }
    else
    {
        mClose->_setFocusHandler(NULL);
        mClose->_setParent(NULL);
        destroy(mClose);
    }
}

bool Window::isResizable()
{
    return mGrip;
}

void Window::close()
{
    setVisible(false);
}

void Window::hide()
{
    const bool oldVisibility = isVisible();

    setVisible(mOldVisibility);

    mOldVisibility = oldVisibility;
}

void Window::mousePressed(gcn::MouseEvent &event)
{
    // Let Guichan move window to top and figure out title bar drag
    gcn::Window::mousePressed(event);

    if (event.getButton() == gcn::MouseEvent::LEFT)
    {
        // Handle window resizing
        mouseResize = getResizeHandles(event);

        if (event.getSource() == this)
            requestFocus();
    }
}

void Window::mouseReleased(gcn::MouseEvent &event)
{
    if (mGrip && mouseResize)
    {
        mouseResize = 0;
        gui->setCursorType(Gui::CURSOR_POINTER);
    }

    // This should be the responsibility of Guichan (and is from 0.8.0 on)
    mMoved = false;
}

void Window::mouseExited(gcn::MouseEvent &event)
{
    if (mGrip && !mouseResize)
        gui->setCursorType(Gui::CURSOR_POINTER);
}

void Window::mouseMoved(gcn::MouseEvent &event)
{
    int resizeHandles = getResizeHandles(event);

    // Changes the custom mouse cursor based on it's current position.
    switch (resizeHandles)
    {
        case BOTTOM | RIGHT:
            gui->setCursorType(Gui::CURSOR_RESIZE_DOWN_RIGHT);
            break;
        case BOTTOM | LEFT:
            gui->setCursorType(Gui::CURSOR_RESIZE_DOWN_LEFT);
            break;
        case BOTTOM:
            gui->setCursorType(Gui::CURSOR_RESIZE_DOWN);
            break;
        case RIGHT:
        case LEFT:
            gui->setCursorType(Gui::CURSOR_RESIZE_ACROSS);
            break;
        default:
            gui->setCursorType(Gui::CURSOR_POINTER);
    }
}

void Window::mouseDragged(gcn::MouseEvent &event)
{
    // Let Guichan handle title bar drag
    gcn::Window::mouseDragged(event);

    // Keep guichan window inside screen when it may be moved
    if (isMovable() && mMoved)
    {
        int newX = std::max(0, getX());
        int newY = std::max(0, getY());
        newX = std::min(graphics->getWidth() - getWidth(), newX);
        newY = std::min(graphics->getHeight() - getHeight(), newY);
        setPosition(newX, newY);
    }

    if (mouseResize && !mMoved)
    {
        const int dx = event.getX() - mDragOffsetX;
        const int dy = event.getY() - mDragOffsetY;
        gcn::Rectangle newDim = getDimension();

        if (mouseResize & (TOP | BOTTOM))
        {
            int newHeight = newDim.height + ((mouseResize & TOP) ? -dy : dy);
            newDim.height = std::min(mMaxWinHeight,
                                     std::max(mMinWinHeight, newHeight));

            if (mouseResize & TOP)
                newDim.y -= newDim.height - getHeight();
        }

        if (mouseResize & (LEFT | RIGHT))
        {
            int newWidth = newDim.width + ((mouseResize & LEFT) ? -dx : dx);
            newDim.width = std::min(mMaxWinWidth,
                                    std::max(mMinWinWidth, newWidth));

            if (mouseResize & LEFT)
                newDim.x -= newDim.width - getWidth();
        }

        // Keep guichan window inside screen (supports resizing any side)
        if (newDim.x < 0)
        {
            newDim.width += newDim.x;
            newDim.x = 0;
        }
        if (newDim.y < 0)
        {
            newDim.height += newDim.y;
            newDim.y = 0;
        }
        if (newDim.x + newDim.width > graphics->getWidth())
        {
            newDim.width = graphics->getWidth() - newDim.x;
        }
        if (newDim.y + newDim.height > graphics->getHeight())
        {
            newDim.height = graphics->getHeight() - newDim.y;
        }

        // Update mouse offset when dragging bottom or right border
        if (mouseResize & BOTTOM)
        {
            mDragOffsetY += newDim.height - getHeight();
        }
        if (mouseResize & RIGHT)
        {
            mDragOffsetX += newDim.width - getWidth();
        }

        // Set the new window and content dimensions
        setDimension(newDim);
    }
}

void Window::loadWindowState()
{
    const std::string &name = mWindowName;
    const std::string skinName = config.getValue(name + "Skin",
                                                 mSkin->getFilePath());
    assert(!name.empty());

    if (mGrip)
    {
        int width = config.getValue(name + "WinWidth", mDefaultWidth);
        int height = config.getValue(name + "WinHeight", mDefaultHeight);

        if (getMinWidth() > width)
            width = getMinWidth();
        else if (getMaxWidth() < width)
            width = getMaxWidth();
        if (getMinHeight() > height)
            height = getMinHeight();
        else if (getMaxHeight() < height)
            height = getMaxHeight();

        setSize(width, height);
    }
    else
    {
        setSize(mDefaultWidth, mDefaultHeight);
    }

    int x = config.getValue(name + "WinX", -1);
    int y = config.getValue(name + "WinY", -1);

    if (x != -1 || y != -1)
    {
        // These two tags are deprecated. Convert them to the new system, then
        // remove them to avoid client coordinate confusion.
        config.removeValue(mWindowName + "WinX");
        config.removeValue(mWindowName + "WinY");
        saveRelativeLocation(x, y);
    }

    int position = config.getValue(name + "Position", -1);
    mOffsetX = config.getValue(name + "OffsetX", mDefaultOffsetX);
    mOffsetY = config.getValue(name + "OffsetY", mDefaultOffsetY);

    if (position != -1)
    {
        mPosition = (ImageRect::ImagePosition) position;
        setLocationRelativeTo(mPosition, mOffsetX, mOffsetY);
    }
    else
    {
        setLocationRelativeTo(mDefaultPosition, mDefaultOffsetX, mDefaultOffsetY);
    }

    if (mSaveVisibility)
    {
        setVisible((bool) config.getValue(name + "Visible", mDefaultVisible));
        mOldVisibility = (bool) config.getValue(name + "Hidden", false);
    }

    if (skinName.compare(mSkin->getFilePath()) != 0)
    {
        mSkin->instances--;
        mSkin = skinLoader->load(skinName, mDefaultSkinPath);
    }
}

void Window::saveWindowState()
{
    // Saving coordinates and Width and Height for resizables in the config
    if (!mWindowName.empty() && mWindowName != "window")
    {
        if (mSaveVisibility)
        {
            config.setValue(mWindowName + "Visible", isVisible());
            config.setValue(mWindowName + "Hidden", mOldVisibility);
        }
        else
        {
            config.removeValue(mWindowName + "Visible");
            config.removeValue(mWindowName + "Hidden");
        }

        config.setValue(mWindowName + "Skin", mSkin->getFilePath());

        saveRelativeLocation(getX(), getY());

        if (mGrip)
        {
            config.setValue(mWindowName + "WinWidth", getWidth());
            config.setValue(mWindowName + "WinHeight", getHeight());
        }
    }
    else if (mWindowName == "window")
    {
        config.removeValue(mWindowName + "Position");
        config.removeValue(mWindowName + "OffsetX");
        config.removeValue(mWindowName + "OffsetY");
    }
}

void Window::setDefaultSize(int defaultWidth, int defaultHeight,
                            ImageRect::ImagePosition position,
                            int offsetX, int offsetY)
{
    if (getMinWidth() > defaultWidth)
        defaultWidth = getMinWidth();
    else if (getMaxWidth() < defaultWidth)
        defaultWidth = getMaxWidth();
    if (getMinHeight() > defaultHeight)
        defaultHeight = getMinHeight();
    else if (getMaxHeight() < defaultHeight)
        defaultHeight = getMaxHeight();

    int posX = 0, posY = 0;

    getRelativeOffset(position, posX, posY, defaultWidth, defaultHeight, 0, 0);

    mDefaultWidth = defaultWidth;
    mDefaultHeight = defaultHeight;
    mDefaultPosition = position;
    mDefaultOffsetX = posX - offsetX;
    mDefaultOffsetY = posY - offsetY;
}

void Window::resetToDefaultSize()
{
    setSize(mDefaultWidth, mDefaultHeight);
    setLocationRelativeTo(mDefaultPosition, mDefaultOffsetX, mDefaultOffsetY);
    saveWindowState();
}

int Window::getResizeHandles(gcn::MouseEvent &event)
{
    int resizeHandles = 0;
    const int y = event.getY();

    if (mGrip && y > (int) mTitleBarHeight)
    {
        const int x = event.getX();

        if (!getChildrenArea().isPointInRect(x, y) && event.getSource() == this)
        {
            resizeHandles |= (x > getWidth() - resizeBorderWidth) ? RIGHT :
                              (x < resizeBorderWidth) ? LEFT : 0;
            resizeHandles |= (y > getHeight() - resizeBorderWidth) ? BOTTOM :
                              (y < resizeBorderWidth) ? TOP : 0;
        }

        if (event.getSource() == mGrip)
        {
            mDragOffsetX = x;
            mDragOffsetY = y;
            resizeHandles |= BOTTOM | RIGHT;
        }
    }

    return resizeHandles;
}

int Window::getGuiAlpha()
{
    float alpha = config.getValue("guialpha", 0.8);
    return (int) (alpha * 255.0f);
}

Layout &Window::getLayout()
{
    if (!mLayout)
        mLayout = new Layout();
    return *mLayout;
}

LayoutCell &Window::place(int x, int y, gcn::Widget *wg, int w, int h)
{
    add(wg);
    return getLayout().place(wg, x, y, w, h);
}

ContainerPlacer Window::getPlacer(int x, int y)
{
    return ContainerPlacer(this, &getLayout().at(x, y));
}

void Window::reflowLayout(int w, int h)
{
    assert(mLayout);
    mLayout->reflow(w, h);
    destroy(mLayout);
    setContentSize(w, h);
}

void Window::refreshLayout()
{
    if (!mLayout)
        return;

    int width = getChildrenArea().width;
    int height = getChildrenArea().height;
    mLayout->reflow(width, height);
}

void Window::fontChanged()
{
    std::list<Widget*>::iterator iter;
    for (iter = mWidgets.begin(); iter != mWidgets.end(); ++iter)
        (*iter)->fontChanged();

    if (!getCaption().empty())
        setTitleBarHeight(getFont()->getHeight() + 10);
}

void Window::clear()
{
    gui->storeFocus();

    gcn::BasicContainer::clear();

    // Restore the resize grip
    if (mGrip)
        add(mGrip);

    if (mClose)
    {
        mClose->_setParent(this);
        mClose->_setFocusHandler(_getFocusHandler());        
    }

    if (mLayout)
        destroy(mLayout);
}

void Window::restoreFocus()
{
    gui->restoreFocus();
}

